/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc-test:fr.bigeon.gclc.test.TestConsoleManager.java
 * Created on: Nov 18, 2016
 */
package fr.bigeon.gclc.manager;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;

/** This console output allows to retrieve the output as an input.
 * <p>
 * This console output is used to internally pilot an application. This can be
 * used to test application behavior.
 *
 * @author Emmanuel Bigeon */
public final class PipedConsoleOutput
        implements ConsoleOutput {

    /** The encoding between streams. */
    private static final String UTF_8 = "UTF-8"; //$NON-NLS-1$
    /** THe inner manager. */
    private final StreamConsoleOutput innerManager;
    /** The reader to get application return from. */
    private final BufferedReader commandBuffOutput;
    /** The stream to get application return from. */
    private final PipedInputStream commandOutput;
    /** The print writer for application to write return to. */
    private final PrintStream outPrint;
    /** The reading thread. */
    private final ReadingRunnable reading;

    /** Create a manager that will write and read through piped stream.
     *
     * @throws IOException if the piping failed for streams */
    public PipedConsoleOutput() throws IOException {
        commandOutput = new PipedInputStream();
        final PipedOutputStream out = new PipedOutputStream(commandOutput);
        commandBuffOutput = new BufferedReader(
                new InputStreamReader(commandOutput, StandardCharsets.UTF_8));
        outPrint = new PrintStream(out, true, UTF_8);
        innerManager = new StreamConsoleOutput(outPrint);
        reading = new ReadingRunnable(commandBuffOutput);
        final Thread th = new Thread(reading, "read"); //$NON-NLS-1$
        th.setDaemon(true);
        th.start();
    }

    /** Test if there is available data.
     *
     * @return the content of the next line written by the application
     * @throws IOException if the reading failed */
    public boolean available() throws IOException {
        return reading.hasMessage();
    }

    /* (non-Javadoc)
     * @see java.lang.AutoCloseable#close() */
    @Override
    public void close() throws IOException {
        reading.setRunning(false);
        innerManager.close();
        outPrint.close();
        commandBuffOutput.close();
        commandOutput.close();
    }

    /** Wait for a specific message to arrive.
     * <p>
     * When this method returns, the message was appended to the data, it
     * <em>may or may not</em> be the next line of data.
     *
     * @param message the message
     * @return the thread to join to wait for message delivery
     * @see fr.bigeon.gclc.manager.ReadingRunnable#getWaitForDelivery(java.lang.String) */
    public Thread getWaitForDelivery(final String message) {
        return reading.getWaitForDelivery(message);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleOutput#isClosed() */
    @Override
    public boolean isClosed() {
        return innerManager.isClosed();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleOutput#print(java.lang.String) */
    @Override
    public void print(final String object) throws IOException {
        innerManager.print(object);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleOutput#println() */
    @Override
    public void println() throws IOException {
        innerManager.println();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleOutput#println(java.lang.String) */
    @Override
    public void println(final String object) throws IOException {
        innerManager.println(object);
    }

    /** Read the next line of data.
     *
     * @return the content of the next line written by the application
     * @throws IOException if the reading failed */
    public String readNextLine() throws IOException {
        return reading.getMessage();
    }
}
